########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

DEFAULT_CLOUDIFY_YAMLLINT_CONFIG = """
extends: default

rules:
  braces:
    max-spaces-inside: 1
    min-spaces-inside: 0
    min-spaces-inside-empty: 0
    max-spaces-inside-empty: 0
  brackets:
    max-spaces-inside: 1
    min-spaces-inside: 0
    min-spaces-inside-empty: 0
    max-spaces-inside-empty: 0
  colons:
    max-spaces-before: 0
    max-spaces-after: 1
  commas:
    max-spaces-after: 1
    max-spaces-before: 0
  comments: disable
  comments-indentation: disable
  document-start:
    present: false
  empty-lines:
    max: 1
  empty-values:
    forbid-in-block-mappings: true
    forbid-in-flow-mappings: true
  hyphens:
    max-spaces-after: 1
  key-duplicates: {}
  indentation:
    spaces: 2
    indent-sequences: consistent
  line-length:
    max: 120
    level: warning
  quoted-strings:
    required: only-when-needed
    extra-allowed: ["^http://", "^ftp://", "(?![A-Za-z0-9])"]
  truthy:
    allowed-values: ['true', 'false']
    check-keys: true
  new-line-at-end-of-file: disable
  new-lines:
    type: platform
  inputs: enable
  imports: enable
  node_types: enable
  dsl_version: enable
  capabilities: enable
  relationships: enable
  node_templates: enable
  dsl_definitions: enable
  blueprint_labels: enable
"""



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cfy_lint.yamllint_ext.utils import context
from contextlib import contextmanager


@contextmanager
def filelines(filename):
    with open(filename, 'r') as file:
        lines = file.readlines()
    yield lines
    with open(filename, 'w') as file:
        file.writelines(lines)


def is_list(line):
    if line.lstrip().startswith('-'):
        return True


def get_eol(line):
    """Gets the end of a line."""
    stripped = line.rstrip()
    eol = ''
    for i in range(0, len(line)):
        try:
            assert stripped[i] == line[i]
        except IndexError:
            if line[i] != ' ':
                eol += line[i]
    return stripped, eol


# Creating a dictionary that connects the changes that happen in
# empty_lined and add_label.
def build_diff_lines():
    if context['add_label'] and not context['line_diff']:
        context['line_diff'][0] = 0
        count = 1
        for i in context['add_label']:
            context['line_diff'].update({i: count})
            count += 1

    elif context['add_label'] and context['line_diff']:
        keys_diff = list(context['line_diff'].keys())
        len_keys_diff = len(keys_diff)
        if context['add_label'][0] < keys_diff[-1]:
            for num in context['add_label']:
                i = 1  # Skip the first member 0
                while i < len_keys_diff and num > keys_diff[i]:
                    i += 1

                k = i
                # We have reached the appropriate index, we will add 1 to it
                while k < len(context['line_diff']):
                    context['line_diff'][keys_diff[k]] = \
                        context['line_diff'][keys_diff[k]] + 1
                    k += 1
                i = k

        # Creating a dict_to_update with the addition of items from add_label
        # Then connecting it with the line_diff dictionary
        i = 0
        while_condition = True
        prev_value = None
        dict_to_update = {}
        for key, value in context['line_diff'].items():
            counter = 0
            while while_condition and key > context['add_label'][i]:
                dict_to_update.update(
                    {context['add_label'][i] + 1: prev_value + 1 + counter})
                i += 1
                counter += 1
                if i >= len(context['add_label']):
                    while_condition = False

            prev_value = value
            if not while_condition:
                break

        while i < len(context['add_label']):
            dict_to_update.update(
                {context['add_label'][i] + 1: prev_value + 1})
            prev_value = prev_value + 1
            i += 1

        context['line_diff'] = connect_two_sorted_dicts(context['line_diff'],
                                                        dict_to_update)


def connect_two_sorted_dicts(dict1, dict2):
    result_dict = {}
    keys1 = iter(dict1)
    keys2 = iter(dict2)
    key1 = next(keys1, None)
    key2 = next(keys2, None)

    while key1 is not None or key2 is not None:
        if key2 is None or (key1 is not None and key1 < key2):
            result_dict[key1] = dict1[key1]
            key1 = next(keys1, None)
        else:
            result_dict[key2] = dict2[key2]
            key2 = next(keys2, None)

    return result_dict



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cfy_lint.yamllint_ext import LintProblem
from cfy_lint.yamllint_ext.generators import CfyNode
from cfy_lint.yamllint_ext.utils import (
    process_relevant_tokens,
    recurse_get_readable_object,
    context as ctx
    )

VALUES = []

ID = 'blueprint_labels'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}
LEVEL0 = 0
LEVEL1 = 1


@process_relevant_tokens(CfyNode, ['blueprint_labels', 'blueprint-labels'])
def check(token=None, **_):
    dsl = ctx.get("dsl_version")
    if dsl == 'cloudify_dsl_1_3':
        yield LintProblem(
            token.prev.line,
            None,
            'cloudify_dsl_1_3 does not support Blueprint Labels')
    if token.prev.node.value == 'blueprint-labels':
        yield LintProblem(
                token.prev.line,
                None,
                'The blueprint_labels key should be written '
                'with an underscore not a dash.')

    for item in token.node.value:
        dictionary = recurse_get_readable_object(item)
        if not isinstance(dictionary, dict):
            yield LintProblem(
                token.line,
                None,
                desc='Every label should be a dictionary')
        else:
            for k, v in dictionary.items():
                if not isinstance(v, dict):
                    yield LintProblem(
                        token.line,
                        None,
                        desc='blueprint_labels contains nested dictionaries',
                        start_mark=item[LEVEL0].start_mark.line,
                        end_mark=item[LEVEL0].end_mark.line)
                else:
                    nested_key = list(v.keys())[LEVEL0]
                    nested_value = list(v.values())[LEVEL0]
                    if nested_key != 'values':
                        yield LintProblem(
                            token.line,
                            None,
                            desc='The name of the key should be "values"',
                            start_mark=item[LEVEL1].start_mark.line,
                            end_mark=item[LEVEL1].end_mark.line)

                    if not isinstance(nested_value, list):
                        non_list_item = item[LEVEL1].value[LEVEL0][LEVEL1]
                        yield LintProblem(
                            token.line,
                            None,
                            'The value of the "values" is should be a list',
                            start_mark=non_list_item.start_mark.line,
                            end_mark=non_list_item.end_mark.line)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yamllint

from .. import generators
from . import get_buffer
from . import get_loader
from . import get_gen_as_list

YAML_CONTENT = """
a:
  aa: aa
  ab: ab
b:
- ba
- bb
c: c
"""


def test_token_or_comment_or_line_generator():
    buffer = get_buffer().read()
    for item in generators.token_or_comment_or_line_generator(buffer):
        if not item:
            continue
        assert isinstance(
            item,
            (generators.CfyNode, generators.CfyToken, yamllint.parser.Line))


def test_generate_nodes_recursively():
    yaml_loader = get_loader(YAML_CONTENT)
    yaml_loader.check_node()
    result = get_gen_as_list(
        generators.generate_nodes_recursively, yaml_loader.get_node().value)
    assert result[7].value == result[-4:-2]



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from mock import Mock
from tempfile import NamedTemporaryFile

from cfy_lint.cli import FixParamValue
from cfy_lint.yamllint_ext import autofix
from cfy_lint.yamllint_ext.autofix import colons
from cfy_lint.yamllint_ext.autofix import commas
from cfy_lint.yamllint_ext.autofix import brackets
from cfy_lint.yamllint_ext.autofix import add_label
from cfy_lint.yamllint_ext.autofix import indentation
from cfy_lint.yamllint_ext.autofix import empty_lines
from cfy_lint.yamllint_ext.overrides import LintProblem
from cfy_lint.yamllint_ext.autofix import trailing_spaces
from cfy_lint.yamllint_ext.autofix import deprecated_node_types
from cfy_lint.yamllint_ext.autofix import deprecated_relationships


def test_indentation_autofix():
    relative_root = os.path.join(os.path.dirname(__file__), 'resources')
    good_blueprint = os.path.join(relative_root, 'blueprint.yaml')
    bad_blueprint = os.path.join(relative_root, 'bad-blueprint.yaml')
    good_content = open(good_blueprint).readlines()
    bad_content = open(bad_blueprint).readlines()
    outfile = NamedTemporaryFile(mode='w', delete=False)
    outfile.writelines(bad_content)
    outfile.close()
    try:
        for line in [26, 35, 70]:
            indentation_problem = LintProblem(
                line=line,
                desc='wrong indentation: expected 6 but found 4',
                rule='indentation',
                column=None,
                file=outfile.name,
            )
            indentation_problem.fixes = [Mock(line=line, rule='indentation')]
            indentation.fix_indentation(problem=indentation_problem)
            if line != 70:
                notfixed = open(outfile.name)
                not_fixed_content = notfixed.readlines()
                notfixed.close()
                assert good_content != not_fixed_content
        fixed = open(outfile.name)
        fixed_content = fixed.readlines()
        fixed.close()
        assert good_content == fixed_content
    finally:
        os.remove(outfile.name)


def test_fix_commas():
    expected = [
        "aws_region_name, z\n",
        "display_label, Aws Region Name\n",
        "type, string\n",
        "default, us-east-1\n",
        "constraints, f\n",
        "valid_values,",
    ]
    lines = """aws_region_name      ,z
display_label     ,     Aws Region Name
type , string
default    , us-east-1
constraints,      f
valid_values ,
"""

    fix_commas_file = get_file(lines)

    try:
        problems = [
            LintProblem(
                line=1,
                column=0,
                desc='too many spaces before colon',
                rule='commas',
                file=fix_commas_file.name
            ),
            LintProblem(
                line=2,
                column=0,
                desc='too many spaces before colon',
                rule='commas',
                file=fix_commas_file.name
            ),
            LintProblem(
                line=3,
                column=0,
                desc='too many spaces before colon',
                rule='commas',
                file=fix_commas_file.name
            ),
            LintProblem(
                line=4,
                column=0,
                desc='too many spaces before colon',
                rule='commas',
                file=fix_commas_file.name
            ),
            LintProblem(
                line=5,
                column=0,
                desc='too many spaces before colon',
                rule='commas',
                file=fix_commas_file.name
            ),
            LintProblem(
                line=6,
                column=0,
                desc='too many spaces before colon',
                rule='commas',
                file=fix_commas_file.name
            ),
        ]
        for problem in problems:
            commas.fix_commas(problem)
    finally:
        f = open(fix_commas_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_commas_file.name)
    assert expected == result_lines


def test_fix_colons():
    expected = [
        "aws_region_name: z\n",
        "display_label: Aws Region Name\n",
        "type: string\n",
        "default: us-east-1\n",
        "constraints: f\n",
        "valid_values: \n",
    ]
    lines = """aws_region_name      :z
display_label     :     Aws Region Name
type : string
default    : us-east-1
constraints:      f
valid_values :
"""

    fix_colons_file = get_file(lines)

    try:
        problems = [
            LintProblem(
                line=1,
                column=0,
                desc='too many spaces before colon',
                rule='colons',
                file=fix_colons_file.name
            ),
            LintProblem(
                line=2,
                column=0,
                desc='too many spaces before colon',
                rule='colons',
                file=fix_colons_file.name
            ),
            LintProblem(
                line=3,
                column=0,
                desc='too many spaces before colon',
                rule='colons',
                file=fix_colons_file.name
            ),
            LintProblem(
                line=4,
                column=0,
                desc='too many spaces before colon',
                rule='colons',
                file=fix_colons_file.name
            ),
            LintProblem(
                line=5,
                column=0,
                desc='too many spaces before colon',
                rule='colons',
                file=fix_colons_file.name
            ),
            LintProblem(
                line=6,
                column=0,
                desc='too many spaces before colon',
                rule='colons',
                file=fix_colons_file.name
            ),
        ]
        for problem in problems:
            colons.fix_colons(problem)
    finally:
        f = open(fix_colons_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_colons_file.name)
    assert expected == result_lines


def get_file(lines):
    fix_truthy_file = NamedTemporaryFile(delete=False)
    f = open(fix_truthy_file.name, 'w')
    if isinstance(lines, str):
        f.write(lines)
    else:
        f.writelines(lines)
    f.close()
    return f


def test_fix_add_label():
    content = """
foo:
  type: bar
baz:
  type: qux
"""
    expected = """
foo:
  display_label: Foo
  type: bar
baz:
  display_label: Baz
  type: qux
"""

    fix_indentation_file = get_file(content)
    problems = [
        LintProblem(
            line=2,
            column=0,
            desc=' is missing a display_label.',
            rule='inputs',
            file=fix_indentation_file.name
        ),
        LintProblem(
            line=4,
            column=0,
            desc=' is missing a display_label.',
            rule='inputs',
            file=fix_indentation_file.name
        ),
    ]
    fix = FixParamValue('all=-1')
    problems[0].fixes = [fix]
    problems[1].fixes = [fix]
    try:
        add_label.fix_add_label(problems)
    finally:
        f = open(fix_indentation_file.name, 'r')
        result = f.read()
        f.close()
        os.remove(fix_indentation_file.name)
    assert expected == result


def test_fix_indentation():
    lines = [
        'foobar:\n',
        '      - foo\n',
        '      - bar\n',
    ]
    expected = [
        'foobar:\n',
        '  - foo\n',
        '  - bar\n',
    ]
    fix_indentation_file = get_file(lines)

    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i,
                column=0,
                desc='wrong indentation: expected 2 but found 6',
                rule='indentation',
                file=fix_indentation_file.name
            )
            indentation.fix_indentation(problem)
    finally:
        f = open(fix_indentation_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_indentation_file.name)
    assert expected == result_lines


def test_braces():
    lines = [
        '{   They wanna get my      } \n',
        '{They wanna get my gold on the ceiling}\n',
        "{     I ain't blind, just a matter of time  }   \n",
        '{ Before you steal it }         \n',
        "{Its all right, ain't no guarding my high   }   \n"
    ]
    expected_lines = [
        '{ They wanna get my } \n',
        '{They wanna get my gold on the ceiling}\n',
        "{ I ain't blind, just a matter of time }   \n",
        '{ Before you steal it }         \n',
        "{Its all right, ain't no guarding my high }   \n"
    ]
    fix_braces_file = get_file(lines)

    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i,
                column=0,
                desc='too many spaces inside braces',
                rule='braces',
                file=fix_braces_file.name
            )
            brackets.fix_spaces_in_brackets(problem)
    finally:
        f = open(fix_braces_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_braces_file.name)
    assert result_lines == expected_lines


def test_brackets():
    lines = [
        '[   They wanna get my      ] \n',
        '[They wanna get my gold on the ceiling]\n',
        "[     I ain't blind, just a matter of time  ]   \n",
        '[ Before you steal it ]         \n',
        "[Its all right, ain't no guarding my high   ]   \n"
    ]
    expected_lines = [
        '[ They wanna get my ] \n',
        '[They wanna get my gold on the ceiling]\n',
        "[ I ain't blind, just a matter of time ]   \n",
        '[ Before you steal it ]         \n',
        "[Its all right, ain't no guarding my high ]   \n"
    ]
    fix_brackets_file = get_file(lines)

    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i,
                column=0,
                desc='too many spaces inside brackets',
                rule='brackets',
                file=fix_brackets_file.name
            )
            brackets.fix_spaces_in_brackets(problem)
    finally:
        f = open(fix_brackets_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_brackets_file.name)
    assert result_lines == expected_lines


def test_trailing_spaces():
    lines = [
        'They wanna get my      \n',
        'They wanna get my gold on the ceiling      \n',
        "I ain't blind, just a matter of time     \n",
        'Before you steal it         \n',
        "Its all right, ain't no guarding my high      \n"
    ]
    expected_lines = [
        'They wanna get my\n',
        'They wanna get my gold on the ceiling\n',
        "I ain't blind, just a matter of time\n",
        'Before you steal it\n',
        "Its all right, ain't no guarding my high\n"
    ]
    fix_trailing_spaces_file = get_file(lines)

    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i,
                column=0,
                desc='foo',
                rule='trailing-spaces',
                file=fix_trailing_spaces_file.name
            )
            trailing_spaces.fix_trailing_spaces(problem)
    finally:
        f = open(fix_trailing_spaces_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_trailing_spaces_file.name)
    assert result_lines == expected_lines


def test_fix_truthy():
    lines = [
        '\n',
        'True, not everything is tRue.\n'
        'And if I say TRUE to you!\n',
        'true TruE TRue TrUE truE\n',
        'False falSe FalsE FALSE false,\n',
        '     "False"      "falsE"\n',
        '\n',
    ]
    expected_lines = [
        '\n',
        'True, not everything is tRue.\n',
        'And if I say true to you!\n',
        'true true true true true\n',
        'false false false false false,\n',
        '     "False"      "falsE"\n',
        '\n'
    ]
    fix_truthy_file = get_file(lines)
    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i + 1,
                column=0,
                desc='foo',
                rule='truthy',
                file=fix_truthy_file.name
            )
            autofix.fix_truthy(problem)
    finally:
        f = open(fix_truthy_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_truthy_file.name)
    assert result_lines == expected_lines


def test_empty_lines():
    lines = [
        "\n",
        "They wanna get my\n",
        "\n",
        "\n",
        "\n",
        "They wanna get my gold on the ceiling\n",
        "I ain't blind, just a matter of time\n",
        "Before you steal it\n",
        "\n",
        "\n",
        "\n",
        "Its all right, ain't no guarding my high\n",
        "\n",
        "\n",
        "\n"
    ]
    expected_lines = [
        "They wanna get my\n",
        "\n",
        "They wanna get my gold on the ceiling\n",
        "I ain't blind, just a matter of time\n",
        "Before you steal it\n",
        "\n",
        "Its all right, ain't no guarding my high\n",
    ]
    fix_empty_lines_file = get_file(lines)

    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i,
                column=0,
                desc='foo',
                rule='trailing-spaces',
                file=fix_empty_lines_file.name
            )
            fix = FixParamValue('all=-1')
            problem.fixes = [fix]
            empty_lines.fix_empty_lines(problem)
    finally:
        f = open(fix_empty_lines_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_empty_lines_file.name)
    assert result_lines == expected_lines


def test_deprecated_node_types():
    lines = [
        'cloudify.azure.nodes.resources.Azure\n',
        'cloudify.azure.nodes.compute.ManagedCluster\n',
        'cloudify.azure.nodes.compute.ContainerService\n',
        'cloudify.azure.nodes.network.LoadBalancer.Probe\n',
        'cloudify.azure.nodes.network.LoadBalancer.BackendAddressPool\n',
        'cloudify.azure.nodes.network.LoadBalancer.IncomingNATRule\n',
        'cloudify.azure.nodes.network.LoadBalancer.Rule\n',
        'cloudify.azure.nodes.network.LoadBalancer\n',
        'cloudify.azure.nodes.compute.VirtualMachineExtension\n',
        'cloudify.azure.nodes.PublishingUser\n',
        'cloudify.azure.nodes.WebApp\n',
        'cloudify.azure.nodes.Plan\n',
        'cloudify.azure.nodes.compute.WindowsVirtualMachine\n',
        'cloudify.azure.nodes.compute.AvailabilitySet\n',
        'cloudify.azure.nodes.network.Route\n',
        'cloudify.azure.nodes.network.NetworkSecurityRule\n',
        'cloudify.azure.nodes.network.RouteTable\n',
        'cloudify.azure.nodes.network.Subnet\n',
        'cloudify.azure.nodes.compute.VirtualMachine\n',
        'cloudify.azure.nodes.network.NetworkInterfaceCard\n',
        'cloudify.azure.nodes.network.NetworkSecurityGroup\n',
        'cloudify.azure.nodes.network.IPConfiguration\n',
        'cloudify.azure.nodes.network.VirtualNetwork\n',
        'cloudify.azure.nodes.network.PublicIPAddress\n',
        'cloudify.azure.nodes.ResourceGroup\n',
        'cloudify.azure.nodes.storage.StorageAccount\n',
        'cloudify.azure.nodes.storage.DataDisk\n',
        'cloudify.azure.nodes.storage.FileShare\n',
        'cloudify.azure.nodes.storage.VirtualNetwork\n',
        'cloudify.azure.nodes.storage.NetworkSecurityGroup\n',
        'cloudify.azure.nodes.storage.NetworkSecurityRule\n',
        'cloudify.azure.nodes.storage.RouteTable\n',
        'cloudify.azure.nodes.storage.Route\n',
        'cloudify.azure.nodes.storage.IPConfiguration\n',
        'cloudify.azure.nodes.storage.PublicIPAddress\n',
        'cloudify.azure.nodes.storage.AvailabilitySet\n',
        'cloudify.azure.nodes.storage.VirtualMachine\n',
        'cloudify.azure.nodes.storage.WindowsVirtualMachine\n',
        'cloudify.azure.nodes.storage.VirtualMachineExtension\n',
        'cloudify.azure.nodes.storage.LoadBalancer\n',
        'cloudify.azure.nodes.storage.BackendAddressPool\n',
        'cloudify.azure.nodes.storage.Probe\n',
        'cloudify.azure.nodes.storage.IncomingNATRule\n',
        'cloudify.azure.nodes.storage.Rule\n',
        'cloudify.azure.nodes.storage.ContainerService\n',
        'cloudify.azure.nodes.storage.Plan\n',
        'cloudify.azure.nodes.storage.WebApp\n',
        'cloudify.azure.nodes.storage.PublishingUser\n',
        'cloudify.azure.nodes.storage.ManagedCluster\n',
        'cloudify.azure.nodes.storage.Azure\n',

        'cloudify.openstack.nodes.Server\n',
        'cloudify.openstack.nodes.WindowsServer\n',
        'cloudify.openstack.nodes.KeyPair\n',
        'cloudify.openstack.nodes.Subnet\n',
        'cloudify.openstack.nodes.SecurityGroup\n',
        'cloudify.openstack.nodes.Router\n',
        'cloudify.openstack.nodes.Port\n',
        'cloudify.openstack.nodes.Network\n',
        'cloudify.openstack.nodes.FloatingIP\n',
        'cloudify.openstack.nodes.RBACPolicy\n',
        'cloudify.openstack.nodes.Volume\n',
        'cloudify.openstack.nova_net.nodes.FloatingIP\n',
        'cloudify.openstack.nova_net.nodes.SecurityGroup\n',
        'cloudify.openstack.nodes.Flavor\n',
        'cloudify.openstack.nodes.Image\n',
        'cloudify.openstack.nodes.Project\n',
        'cloudify.openstack.nodes.User\n',
        'cloudify.openstack.nodes.HostAggregate\n',
        'cloudify.openstack.nodes.ServerGroup\n',
        'cloudify.openstack.nodes.Routes\n'
    ]

    expected_lines = [
        'cloudify.nodes.azure.resources.Azure\n',
        'cloudify.nodes.azure.compute.ManagedCluster\n',
        'cloudify.nodes.azure.compute.ContainerService\n',
        'cloudify.nodes.azure.network.LoadBalancer.Probe\n',
        'cloudify.nodes.azure.network.LoadBalancer.BackendAddressPool\n',
        'cloudify.nodes.azure.network.LoadBalancer.IncomingNATRule\n',
        'cloudify.nodes.azure.network.LoadBalancer.Rule\n',
        'cloudify.nodes.azure.network.LoadBalancer\n',
        'cloudify.nodes.azure.compute.VirtualMachineExtension\n',
        'cloudify.nodes.azure.PublishingUser\n',
        'cloudify.nodes.azure.WebApp\n',
        'cloudify.nodes.azure.Plan\n',
        'cloudify.nodes.azure.compute.WindowsVirtualMachine\n',
        'cloudify.nodes.azure.compute.AvailabilitySet\n',
        'cloudify.nodes.azure.network.Route\n',
        'cloudify.nodes.azure.network.NetworkSecurityRule\n',
        'cloudify.nodes.azure.network.RouteTable\n',
        'cloudify.nodes.azure.network.Subnet\n',
        'cloudify.nodes.azure.compute.VirtualMachine\n',
        'cloudify.nodes.azure.network.NetworkInterfaceCard\n',
        'cloudify.nodes.azure.network.NetworkSecurityGroup\n',
        'cloudify.nodes.azure.network.IPConfiguration\n',
        'cloudify.nodes.azure.network.VirtualNetwork\n',
        'cloudify.nodes.azure.network.PublicIPAddress\n',
        'cloudify.nodes.azure.ResourceGroup\n',
        'cloudify.nodes.azure.storage.StorageAccount\n',
        'cloudify.nodes.azure.storage.DataDisk\n',
        'cloudify.nodes.azure.storage.FileShare\n',
        'cloudify.nodes.azure.storage.VirtualNetwork\n',
        'cloudify.nodes.azure.storage.NetworkSecurityGroup\n',
        'cloudify.nodes.azure.storage.NetworkSecurityRule\n',
        'cloudify.nodes.azure.storage.RouteTable\n',
        'cloudify.nodes.azure.storage.Route\n',
        'cloudify.nodes.azure.storage.IPConfiguration\n',
        'cloudify.nodes.azure.storage.PublicIPAddress\n',
        'cloudify.nodes.azure.storage.AvailabilitySet\n',
        'cloudify.nodes.azure.storage.VirtualMachine\n',
        'cloudify.nodes.azure.storage.WindowsVirtualMachine\n',
        'cloudify.nodes.azure.storage.VirtualMachineExtension\n',
        'cloudify.nodes.azure.storage.LoadBalancer\n',
        'cloudify.nodes.azure.storage.BackendAddressPool\n',
        'cloudify.nodes.azure.storage.Probe\n',
        'cloudify.nodes.azure.storage.IncomingNATRule\n',
        'cloudify.nodes.azure.storage.Rule\n',
        'cloudify.nodes.azure.storage.ContainerService\n',
        'cloudify.nodes.azure.storage.Plan\n',
        'cloudify.nodes.azure.storage.WebApp\n',
        'cloudify.nodes.azure.storage.PublishingUser\n',
        'cloudify.nodes.azure.storage.ManagedCluster\n',
        'cloudify.nodes.azure.storage.Azure\n',

        'cloudify.nodes.openstack.Server\n',
        'cloudify.nodes.openstack.WindowsServer\n',
        'cloudify.nodes.openstack.KeyPair\n',
        'cloudify.nodes.openstack.Subnet\n',
        'cloudify.nodes.openstack.SecurityGroup\n',
        'cloudify.nodes.openstack.Router\n',
        'cloudify.nodes.openstack.Port\n',
        'cloudify.nodes.openstack.Network\n',
        'cloudify.nodes.openstack.FloatingIP\n',
        'cloudify.nodes.openstack.RBACPolicy\n',
        'cloudify.nodes.openstack.Volume\n',
        'cloudify.nodes.openstack.FloatingIP\n',
        'cloudify.nodes.openstack.SecurityGroup\n',
        'cloudify.nodes.openstack.Flavor\n',
        'cloudify.nodes.openstack.Image\n',
        'cloudify.nodes.openstack.Project\n',
        'cloudify.nodes.openstack.User\n',
        'cloudify.nodes.openstack.HostAggregate\n',
        'cloudify.nodes.openstack.ServerGroup\n',
        'cloudify.nodes.openstack.Router\n'
    ]

    fix_trailing_spaces_file = get_file(lines)

    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i+1,
                column=0,
                desc='deprecated node type. Replace usage of {} with '
                     '{}'.format(lines[i], expected_lines[i]),
                rule='node_templates',
                file=fix_trailing_spaces_file.name,
                fixable=True
            )
            deprecated_node_types.fix_deprecated_node_types(problem)
    finally:
        f = open(fix_trailing_spaces_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_trailing_spaces_file.name)
    assert result_lines == expected_lines


def test_relationships_types():
    lines = [
        'cloudify.azure.relationships.contained_in_resource_group\n',
        'cloudify.azure.relationships.contained_in_storage_account\n',
        'cloudify.azure.relationships.contained_in_virtual_network\n',
        'cloudify.azure.relationships.contained_in_network_security_group\n',
        'cloudify.azure.relationships.contained_in_route_table\n',
        'cloudify.azure.relationships.contained_in_load_balancer\n',
        'cloudify.azure.relationships.'
        'network_security_group_attached_to_subnet\n',
        'cloudify.azure.relationships.route_table_attached_to_subnet\n',
        'cloudify.azure.relationships.nic_connected_to_ip_configuration\n',
        'cloudify.azure.relationships.ip_configuration_connected_to_subnet\n',
        'cloudify.azure.relationships.'
        'ip_configuration_connected_to_public_ip\n',
        'cloudify.azure.relationships.connected_to_storage_account\n',
        'cloudify.azure.relationships.connected_to_data_disk\n',
        'cloudify.azure.relationships.connected_to_nic\n',
        'cloudify.azure.relationships.connected_to_availability_set\n',
        'cloudify.azure.relationships.connected_to_ip_configuration\n',
        'cloudify.azure.relationships.connected_to_lb_be_pool\n',
        'cloudify.azure.relationships.connected_to_lb_probe\n',
        'cloudify.azure.relationships.vmx_contained_in_vm\n',
        'cloudify.azure.relationships.nic_connected_to_lb_be_pool\n',
        'cloudify.azure.relationships.vm_connected_to_datadisk\n',
        'cloudify.azure.relationships.connected_to_aks_cluster\n',

        'cloudify.openstack.server_connected_to_server_group\n',
        'cloudify.openstack.server_connected_to_keypair\n',
        'cloudify.openstack.server_connected_to_port\n',
        'cloudify.openstack.server_connected_to_floating_ip\n',
        'cloudify.openstack.server_connected_to_security_group\n',
        'cloudify.openstack.port_connected_to_security_group\n',
        'cloudify.openstack.port_connected_to_floating_ip\n',
        'cloudify.openstack.port_connected_to_subnet\n',
        'cloudify.openstack.subnet_connected_to_router\n',
        'cloudify.openstack.volume_attached_to_server\n',
        'cloudify.openstack.route_connected_to_router\n',
        'cloudify.openstack.rbac_policy_applied_to\n'
    ]

    expected_lines = [
        'cloudify.relationships.azure.contained_in_resource_group\n',
        'cloudify.relationships.azure.contained_in_storage_account\n',
        'cloudify.relationships.azure.contained_in_virtual_network\n',
        'cloudify.relationships.azure.contained_in_network_security_group\n',
        'cloudify.relationships.azure.contained_in_route_table\n',
        'cloudify.relationships.azure.contained_in_load_balancer\n',
        'cloudify.relationships.azure.network_security_group_attached_'
        'to_subnet\n',
        'cloudify.relationships.azure.route_table_attached_to_subnet\n',
        'cloudify.relationships.azure.nic_connected_to_ip_configuration\n',
        'cloudify.relationships.azure.ip_configuration_connected_to_subnet\n',
        'cloudify.relationships.azure.ip_configuration_connected_'
        'to_public_ip\n',
        'cloudify.relationships.azure.connected_to_storage_account\n',
        'cloudify.relationships.azure.connected_to_data_disk\n',
        'cloudify.relationships.azure.connected_to_nic\n',
        'cloudify.relationships.azure.connected_to_availability_set\n',
        'cloudify.relationships.azure.connected_to_ip_configuration\n',
        'cloudify.relationships.azure.connected_to_lb_be_pool\n',
        'cloudify.relationships.azure.connected_to_lb_probe\n',
        'cloudify.relationships.azure.vmx_contained_in_vm\n',
        'cloudify.relationships.azure.nic_connected_to_lb_be_pool\n',
        'cloudify.relationships.azure.vm_connected_to_datadisk\n',
        'cloudify.relationships.azure.connected_to_aks_cluster\n',

        'cloudify.relationships.openstack.server_connected_to_server_group\n',
        'cloudify.relationships.openstack.server_connected_to_keypair\n',
        'cloudify.relationships.openstack.server_connected_to_port\n',
        'cloudify.relationships.openstack.server_connected_to_floating_ip\n',
        'cloudify.relationships.openstack.server_connected_'
        'to_security_group\n',
        'cloudify.relationships.openstack.port_connected_to_security_group\n',
        'cloudify.relationships.openstack.port_connected_to_floating_ip\n',
        'cloudify.relationships.openstack.port_connected_to_subnet\n',
        'cloudify.relationships.openstack.subnet_connected_to_router\n',
        'cloudify.relationships.openstack.volume_attached_to_server\n',
        'cloudify.relationships.openstack.route_connected_to_router\n',
        'cloudify.relationships.openstack.rbac_policy_applied_to\n'
    ]

    fix_trailing_spaces_file = get_file(lines)

    try:
        for i in range(0, len(lines)):
            problem = LintProblem(
                line=i,
                column=0,
                desc='deprecated relationship type. Replace usage '
                     'of {} with {}'.format(lines[i], expected_lines[i]),
                rule='relationships',
                file=fix_trailing_spaces_file.name,
                fixable=True
            )
            deprecated_relationships.fix_deprecated_relationships(problem)
    finally:
        f = open(fix_trailing_spaces_file.name, 'r')
        result_lines = f.readlines()
        f.close()
        os.remove(fix_trailing_spaces_file.name)

    assert result_lines == expected_lines


def test_client_config():
    lines = [
        '  vm:\n',
        '    type: cloudify.nodes.aws.ec2.Instances\n',
        '    properties:\n',
        '      aws_config: *client_config\n',
    ]
    expected_lines = [
        '  vm:\n',
        '    type: cloudify.nodes.aws.ec2.Instances\n',
        '    properties:\n',
        '      client_config: *client_config\n'
    ]
    fix_clinet_config_file = get_file(lines)

    problem = LintProblem(
        line=3,
        column=0,
        desc='The node template "vm" has deprecated property '
             '"aws_config". please use "client_config"',
        rule='node_templates',
        file=fix_clinet_config_file.name,
        fixable=True
    )
    deprecated_node_types.fix_deprecated_node_types(problem)
    f = open(fix_clinet_config_file.name, 'r')
    result_lines = f.readlines()
    f.close()
    os.remove(fix_clinet_config_file.name)
    assert result_lines == expected_lines



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import re
import sys
import json
import urllib
from re import sub
from logging import (Formatter, StreamHandler)

from cfy_lint import cli, __version__
from cfy_lint.yamllint_ext import LintProblem
from cfy_lint.yamllint_ext import (run, rules)
from cfy_lint.logger import logger, stream_handler
from cfy_lint.yamllint_ext.autofix import fix_problem
from cfy_lint.yamllint_ext.config import YamlLintConfigExt
from cfy_lint.yamllint_ext.autofix.add_label import fix_add_label
from cfy_lint.yamllint_ext.autofix.empty_lines import fix_empty_lines


def report_both_fix_autofix(af, f):
    f = f or []
    if af and f:
        print('The parameters -af/--autofix and --fix are '
              'mutually exclusive. Use --help for more info.')
        sys.exit(1)
    elif af:
        f.insert(0, cli.FixParamValue('all=-1'))
    return f


def format_json(format):
    if format == 'json':
        logger.removeHandler(stream_handler)
        new_logging_handler = StreamHandler()
        new_logging_formatter = Formatter(fmt='%(message)s')
        new_logging_handler.setFormatter(new_logging_formatter)
        logger.addHandler(new_logging_handler)


@cli.command()
@cli.options.blueprint_path
@cli.options.config
@cli.options.verbose
@cli.options.format
@cli.options.skip_suggestions
@cli.options.autofix
@cli.options.fix
@cli.options.fix_only
@cli.click.version_option(__version__.version)
def lint(blueprint_path,
         config,
         verbose,
         format,
         skip_suggestions=None,
         autofix=False,
         fix=None,
         fix_only=None,
         **_):

    if fix_only:
        extra_empty_line = False
        add_label_offset = False
        problems = []
        for x in fix_only:
            x = json.loads(x)
            problem = LintProblem(
                line=x['line'],
                column=None,
                desc=x['message'],
                rule=x['rule'],
            )
            input_file_path = os.path.abspath(blueprint_path)
            problem.file = input_file_path
            # print(problems)
            if problem.rule == 'inputs' and \
                    'is missing a display_label' in problem.message:
                add_label_offset = True
                problems.append(problem)
            elif problem.rule == 'empty-lines':
                extra_empty_line = True
            else:
                fix_problem(problem)

        if add_label_offset:
            fix_add_label(problems, fix_only=True)

        if extra_empty_line:
            fix_empty_lines(problem)

        sys.exit(0)

    fix = report_both_fix_autofix(autofix, fix)
    format_json(format)

    yaml_config = YamlLintConfigExt(content=config, yamllint_rules=rules)
    skip_suggestions = skip_suggestions or ()
    try:
        report = create_report_for_file(
            blueprint_path,
            yaml_config,
            skip_suggestions=skip_suggestions,
            fix=fix)
    except Exception as e:
        if verbose:
            raise e
        else:
            exception_str = str(e)
        logger.error(exception_str)
        sys.exit(1)

    cnt = 0
    try:
        for item in report:
            message = formatted_message(item, format)
            if cnt == 0:
                logger.info('The following linting errors were found: ')
                cnt += 1
            if item.level == 'warning':
                logger.warning(message)
            elif item.level == 'error':
                logger.error(message)
            else:
                logger.info(message)
    except urllib.error.URLError as e:
        if verbose:
            raise e
        else:
            exception_str = str(e)
        logger.error(exception_str)
        sys.exit(1)


def create_report_for_file(file_path,
                           conf,
                           create_report_for_file=False,
                           skip_suggestions=None,
                           fix=None):
    if not os.path.exists(file_path):
        raise RuntimeError('File path does not exist: {}.'.format(file_path))
    logger.info('Linting blueprint: {}'.format(file_path))
    with io.open(file_path, newline='') as f:
        return run(f, conf, create_report_for_file, skip_suggestions, fix)


def formatted_message(item, format=None):
    if format == 'json':
        rule, item_message = item.message.split(':', 1)
        try:
            result = re.split(r'available\),\sSeverity:\s', item_message)
            if len(result) == 2:
                severity = 2
            else:
                raise Exception('Something is up with that line.')
        except Exception:
            severity = '0'
        return json.dumps({
            "level": item.level,
            "line": item.line,
            "rule": sub(r"[()]", "", rule),
            "message": item_message,
            "severity": int(severity),
        })
    return '{0: <4}: {1:>4}'.format(item.line, item.message)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

af = """Fix all (supported autofix) issues in place.
This is the equivalent of --fix all=-1.
Not all issues can be solved automatically."""

bp = """Path to the blueprint file that you want to lint."""

c = """A path to your own yamllint config file."""

f = """Toggle the cfy-lint output format. Currently default or "-f json"."""

fix = """Fix a single issue in place.
The format is key=value, where key is issue type and value is line number,
e.g. "--fix inputs:21". The "--fix" flag can be used multiple times.
Not all issues can be solved automatically."""

v = """Show verbose output, including exceptions."""

xs = """Do not display suggested values for supported sections."""

fo = """Fix a single issue without linting again.
The format is {"level":"value","line":value,"rule":"value","message":"value"},
where value is a string except in line where its the line number,
e.g. {"level":"error","line":9,"rule":"inputs","message":"Input
autosubnets is missing a display_label."}.
The "--fix-only" flag can be used multiple times.
Not all issues can be solved automatically."""



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from cfy_lint.yamllint_ext.autofix.utils import filelines, get_eol


def fix_deprecated_node_types(problem):
    if problem.rule != 'node_templates':
        return
    if 'deprecated node type' in problem.message:
        with filelines(problem.file) as lines:
            line = lines[problem.line - 1]
            line, eol = get_eol(line)
            split = problem.message.split()
            new_line = line.replace(split[-5], split[-3].rstrip('.'))
            lines[problem.line - 1] = new_line + eol
        problem.fixed = True
    if 'has deprecated property' in problem.message:
        words = problem.message.split()
        pattern = re.compile('(azure|aws|gcp)_config')
        target = words[-6].replace('"', '').replace('.', '') + ":"
        if not pattern.search(target):
            return
        line_number = problem.line - 1
        with filelines(problem.file) as lines:
            while target not in lines[line_number]:
                line_number += 1
            line = lines[line_number]
            line, eol = get_eol(line)
            new_line = re.sub(target, "client_config:", line)
            lines[line_number] = new_line + eol
        problem.fixed = True



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from cfy_lint.yamllint_ext.autofix.utils import filelines, get_eol


def fix_colons(problem):
    if problem.rule == 'colons':
        with filelines(problem.file) as lines:
            line = lines[problem.line - 1]
            if re.match(r'^\s*([^:\n]+)\s*:\s*([^:\n]+)\s*$', line):
                new_line = re.sub(r'\s*:\s*', ': ', line)
            elif re.match(r'^\s*([^:\n]+)\s*:\s*$', line):
                new_line = re.sub(r'\s*:', ': ', line)
            striped_new_line, eol = get_eol(new_line)
            if new_line[-1] != eol:
                new_line = striped_new_line + eol
            lines[problem.line - 1] = new_line
        problem.fixed = True



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from cfy_lint import helptexts

CLICK_CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help'])


class FixParamType(click.types.StringParamType):
    pass


class FixParamValue(object):
    def __init__(self, value):
        if not isinstance(value, str):
            raise TypeError()
        try:
            rule, line = value.split('=')
        except ValueError:
            raise ValueError(
                'Fix must be in the format rule=lineno.')
        self.line = int(line)
        self.rule = rule


def get_fix(ctx, *args, **_):
    fixes = []
    if not args or \
            isinstance(args[-1], click.Option) or \
            len(args[-1]) < 1:
        return
    for item in args[-1]:
        try:
            fixes.append(FixParamValue(item))
        except (TypeError, ValueError) as e:
            print('Invalid parameter for --fix.')
            if hasattr(e, 'message'):
                print(e.message)
            else:
                print(str(e))
            ctx.abort()
    return fixes


def init():
    pass


def group(name):
    return click.group(name=name, context_settings=CLICK_CONTEXT_SETTINGS)


def command(*args, **kwargs):
    return click.command(*args, **kwargs)


class Options(object):
    def __init__(self):

        self.blueprint_path = click.option(
            '-b',
            '--blueprint-path',
            default='blueprint.yaml',
            type=click.Path(),
            multiple=False,
            show_default='blueprint.yaml',
            help=helptexts.bp)

        self.config = click.option(
            '-c',
            '--config',
            default=None,
            type=click.Path(),
            multiple=False,
            help=helptexts.c)

        self.verbose = click.option(
            '-v',
            '--verbose',
            default=False,
            type=click.BOOL,
            is_flag=True,
            multiple=False,
            help=helptexts.v)

        self.format = click.option(
            '-f',
            '--format',
            default=None,
            type=click.STRING,
            multiple=False,
            help=helptexts.f)

        self.skip_suggestions = click.option(
            '-xs',
            '--skip-suggestions',
            default=None,
            type=click.STRING,
            multiple=True,
            help=helptexts.xs)

        self.autofix = click.option(
            '-af',
            '--autofix',
            default=False,
            type=click.BOOL,
            is_flag=True,
            multiple=False,
            help=helptexts.af)

        self.fix = click.option(
            '--fix',
            type=FixParamType(),
            callback=get_fix,
            multiple=True,
            help=helptexts.fix)

        self.fix_only = click.option(
            '-fo',
            '--fix-only',
            default=None,
            type=click.STRING,
            multiple=True,
            help=helptexts.fo)


options = Options()



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml


def spaces_after(token, prev, next, min=-1, max=-1,
                 min_desc=None, max_desc=None):
    if next is not None and token.end_mark.line == next.start_mark.line:
        spaces = next.start_mark.pointer - token.end_mark.pointer
        if max != - 1 and spaces > max:
            return LintProblem(token.start_mark.line + 1,
                               next.start_mark.column,
                               max_desc,
                               fixable=True)
        elif min != - 1 and spaces < min:
            return LintProblem(token.start_mark.line + 1,
                               next.start_mark.column + 1,
                               min_desc,
                               fixable=True)


def spaces_before(token, prev, next, min=-1, max=-1,
                  min_desc=None, max_desc=None):
    if (prev is not None and prev.end_mark.line == token.start_mark.line and
            # Discard tokens (only scalars?) that end at the start of next line
            (prev.end_mark.pointer == 0 or
             prev.end_mark.buffer[prev.end_mark.pointer - 1] != '\n')):
        spaces = token.start_mark.pointer - prev.end_mark.pointer
        if max != - 1 and spaces > max:
            return LintProblem(token.start_mark.line + 1,
                               token.start_mark.column,
                               max_desc,
                               fixable=True)
        elif min != - 1 and spaces < min:
            return LintProblem(token.start_mark.line + 1,
                               token.start_mark.column + 1,
                               min_desc,
                               fixable=True)


class LintProblem(object):
    """Represents a linting problem found by yamllint."""
    def __init__(self,
                 line,
                 column,
                 desc='<no description>',
                 rule=None,
                 file=None,
                 token=None,
                 start_mark=None,
                 end_mark=None,
                 next=None,
                 prev=None,
                 nextnext=None,
                 fixable=None,
                 severity=None):
        #: Line on which the problem was found (starting at 1)
        self._line = line
        #: Column on which the problem was found (starting at 1)
        self.column = column or 0
        #: Human-readable description of the problem
        self._desc = desc
        #: Identifier of the rule that detected the problem
        self.rule = rule
        self.level = None
        self._file = file
        self._token = token
        self._start_mark = start_mark
        self._end_mark = end_mark
        self._next = next
        self._prev = prev
        self._nextnext = nextnext
        self._fixed = False
        self._fixes = []
        self._fix = False
        self._fixable = fixable
        self._update_line = None
        self._severity = severity

    @property
    def line(self):
        if self.update_line:
            return self.update_line
        if self.start_mark and self.end_mark:
            return self.start_mark + 1
        else:
            return self._line

    @line.setter
    def line(self, value):
        self._line = value

    @property
    def update_line(self):
        return self._update_line

    @update_line.setter
    def update_line(self, value):
        self._update_line = value

    @property
    def desc(self):
        if self.fixable:
            fixablity = " (auto-fix available)"
        else:
            fixablity = " (auto-fix unavailable)"
        if self.severity:
            severity_message = ', Severity: {}'.format(self.severity)
            return self._desc + fixablity + severity_message
        else:
            return self._desc + fixablity

    @property
    def severity(self):
        return self._severity

    @severity.setter
    def severity(self, value):
        self._severity = value

    @property
    def fixable(self):
        return self._fixable

    @fixable.setter
    def fixable(self, value):
        self._fixable = value

    @property
    def fixes(self):
        return self._fixes

    @fixes.setter
    def fixes(self, value):
        self._fixes = value

    @property
    def fix_all(self):
        for fix in self.fixes:
            if fix.line == -1 and fix.rule == 'all':
                return True
        return False

    @property
    def fix(self):
        for fix in self.fixes:
            if fix.line == -1 and fix.rule == 'all':
                return True
            elif fix.line == self.line and fix.rule == self.rule:
                return True
        return False

    @property
    def fix_new_lines(self):
        for fix in self.fixes:
            if fix.rule == 'empty-lines':
                return True
        return False

    @property
    def fixed(self):
        return self._fixed

    @fixed.setter
    def fixed(self, value):
        self._fixed = value

    @property
    def message(self):
        if self.rule is not None:
            return '({}): {}'.format(self.rule, self.desc)
        return self.desc

    def __eq__(self, other):
        return (self.line == other.line and
                self.column == other.column and
                self.rule == other.rule)

    def __lt__(self, other):
        return (self.line < other.line or
                (self.line == other.line and self.column < other.column))

    def __repr__(self):
        return '%d:%d: %s' % (self.line, self.column, self.message)

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, value):
        self._file = value

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, value):
        self._token = value

    @property
    def next(self):
        return self._next

    @next.setter
    def next(self, value):
        self._next = value

    @property
    def start_mark(self):
        return self._start_mark

    @start_mark.setter
    def start_mark(self, value):
        self._start_mark = value

    @property
    def end_mark(self):
        return self._end_mark

    @end_mark.setter
    def end_mark(self, value):
        self._end_mark = value

    @property
    def nextnext(self):
        return self._nextnext

    @nextnext.setter
    def nextnext(self, value):
        self._nextnext = value

    @property
    def prev(self):
        return self._prev

    @prev.setter
    def prev(self, value):
        self._prev = value


def get_syntax_error(buffer):
    try:
        list(yaml.parse(buffer, Loader=yaml.BaseLoader))
    except yaml.error.MarkedYAMLError as e:
        problem = LintProblem(e.problem_mark.line + 1,
                              e.problem_mark.column + 1,
                              'syntax error: ' + e.problem + ' (syntax)')
        problem.level = 'error'
        return problem



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

from cfy_lint.yamllint_ext.utils import context
from cfy_lint.yamllint_ext.autofix.utils import filelines

TYP = 'inputs'
MSG = r'is\smissing\sa\sdisplay_label'
INDENT = r'^\s+'
INDENT_EMPTY_LINES = r'^ \s+'
EMPTY = r'^\s*$'


def fix_add_label(problems, fix_only=False):
    counter = 0
    for problem in problems:
        if not problem.fix and not fix_only:
            continue
        if problem.rule == TYP and re.search(MSG, problem.message):
            with filelines(problem.file) as lines:
                label = lines[problem.line - 1 + counter]
                try:
                    is_empty_line = re.findall(EMPTY,
                                               lines[problem.line + counter])
                    if is_empty_line:
                        while not re.findall(EMPTY,
                                             lines[problem.line + counter]):
                            counter += 1
                        indentation = re.search(
                            INDENT_EMPTY_LINES,
                            lines[problem.line + counter]).group()
                    else:
                        indentation = re.search(
                            INDENT,
                            lines[problem.line + counter]).group()
                except AttributeError:
                    indentation = ''.join(
                        re.search(INDENT, label).group() +
                        re.search(INDENT, label).group()
                    )

                label = label.strip().replace('_', ' ').replace(':', '')
                label = '{indentation}display_label: ' \
                        '{label}{linesep}'.format(
                            indentation=indentation,
                            label=label.title(),
                            linesep='\n')
                lines.insert(problem.line + counter, label)
                counter += 1
                context['add_label'].append(problem.line)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cfy_lint.yamllint_ext import LintProblem
from cfy_lint.yamllint_ext.rules import constants
from cfy_lint.yamllint_ext.generators import CfyNode
from cfy_lint.yamllint_ext.utils import (
    process_relevant_tokens,
    check_node_imported,
    recurse_get_readable_object,
    context as ctx
    )
from cfy_lint.yamllint_ext.rules.node_templates import (
    remove_node_type_from_context
)

VALUES = []

ID = 'node_types'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}


@process_relevant_tokens(CfyNode, 'node_types')
def check(token=None, skip_suggestions=None, **_):
    for node_type in token.node.value:
        types = get_type_and_check_dsl(node_type)
        dsl = ctx.get("dsl_version")
        for value in types:
            if value not in constants.INPUTS_BY_DSL.get(dsl, []):
                yield LintProblem(
                    token.line,
                    None,
                    'Type {} is not supported by DSL {}.'.format(value, dsl)
                )
        if check_node_imported(node_type[0].value):
            yield from node_type_follows_naming_conventions(
                node_type[0].value, token.line, skip_suggestions)
    remove_node_type_from_context(node_type)


def get_values_by_key_type(dictionary):
    values = []
    if 'type' in dictionary:
        values.append(dictionary['type'])
    for value in dictionary.values():
        if isinstance(value, dict):
            nested_values = get_values_by_key_type(value)
            values.extend(nested_values)
    return values


def get_type_and_check_dsl(node_type):
    node_type = recurse_get_readable_object(node_type)
    return get_values_by_key_type(node_type)


def node_type_follows_naming_conventions(value, line, skip_suggestions=None):
    suggestions = 'node_templates' in skip_suggestions
    split_node_type = value.split('.')
    last_key = split_node_type.pop()
    if not {'cloudify', 'nodes'} <= set(split_node_type):
        yield LintProblem(
            line,
            None,
            "node types should follow naming convention cloudify.nodes.*: "
            "{}".format(value))
    if not good_camel_case(last_key, split_node_type) and not suggestions:
        new_value = '.'.join(
            [k.lower() for k in split_node_type]) + '.{}'.format(last_key)
        yield LintProblem(
            line,
            None,
            "incorrect camel case {}. Suggested: {} ".format(value, new_value))


def good_camel_case(last_key, split_node_type):
    if not last_key[0].isupper():
        return False
    for key in split_node_type:
        if key[0].isupper():
            return False
    return True



########
# Copyright (c) 2014-2023 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

from cfy_lint.yamllint_ext import LintProblem
from cfy_lint.yamllint_ext.rules import constants
from cfy_lint.yamllint_ext.generators import CfyNode
from cfy_lint.yamllint_ext.constants import UNUSED_INPUTS
from cfy_lint.yamllint_ext.utils import (
    INTRINSIC_FNS,
    recurse_mapping,
    context as ctx, process_relevant_tokens)

VALUES = []

ID = 'inputs'
TYPE = 'token'
CONF = {'allowed-values': list(VALUES), 'check-keys': bool}
DEFAULT = {'allowed-values': ['true', 'false'], 'check-keys': True}
ALLOWED_KEYS = [
    'type',
    'hidden',
    'default',
    'required',
    'constraints',
    'description',
    'display_label'
]
STR_INTRINSIC_FNS = [
    'concat',
    'string_lower',
    'string_upper',
    'string_replace'
]
DICT_INTRINSIC_FNS = [
    'merge'
]
INT_INTRINSIC_FNS = [
    'string_find'
]
LEVEL0 = 0
LEVEL1 = 1


@process_relevant_tokens(CfyNode, ['inputs', 'get_input'])
def check(token=None, skip_suggestions=None, **_):
    if token.prev.node.value == 'inputs':
        for item in token.node.value:
            if isinstance(item, yaml.nodes.ScalarNode):
                yield LintProblem(
                    token.line,
                    None,
                    'Bad inputs format. '
                    'Input should be a key not a list item.')
            input_obj = CfyInput(item)
            if not input_obj.name and not input_obj.mapping:
                continue
            if input_obj.not_input() and get_default_type(input_obj) != bool:
                continue
            ctx['inputs'].update(input_obj.__dict__())
            if input_obj.name not in ctx[UNUSED_INPUTS]:
                input_unused = item[LEVEL0]
                ctx[UNUSED_INPUTS].update(
                    {
                        input_obj.name: LintProblem(
                            token.line,
                            None,
                            desc='input {} is unused.'.format(input_obj.name),
                            start_mark=input_unused.start_mark.line,
                            end_mark=input_unused.end_mark.line
                        )
                    }
                )
            yield from validate_inputs(input_obj,
                                       input_obj.line or token.line,
                                       ctx.get("dsl_version"),
                                       skip_suggestions,
                                       item)

    if token.prev.node.value == 'get_input':
        if isinstance(token.node.value, list):
            if isinstance(token.node.value[0], yaml.nodes.ScalarNode):
                if token.node.value[0].value not in ctx['inputs']:
                    yield LintProblem(
                        token.line,
                        None,
                        'undefined input {}'
                        .format(token.node.value[0].value))
                elif token.node.value[0].value in ctx[UNUSED_INPUTS]:
                    del ctx[UNUSED_INPUTS][token.node.value[0].value]
            if isinstance(token.node.value[0], tuple):
                if token.node.value[0][0] not in ctx['inputs']:
                    yield LintProblem(
                        token.line,
                        None,
                        'undefined input "{}"'.format(token.node.value[0][0]))
                elif token.node.value[0][0] in ctx[UNUSED_INPUTS]:
                    del ctx[UNUSED_INPUTS][token.node.value[0][0]]
        else:
            if token.node.value not in ctx['inputs']:
                yield LintProblem(
                    token.line,
                    None,
                    'undefined input "{}"'.format(token.node.value))
            elif token.node.value in ctx[UNUSED_INPUTS]:
                del ctx[UNUSED_INPUTS][token.node.value]


def validate_inputs(input_obj, line, dsl, skip_suggestions=None, item=None):
    suggestions = 'inputs' in skip_suggestions
    if input_obj.invalid_keys:
        yield LintProblem(
            line,
            None,
            'the following keys are invalid for inputs: {}'.format(
                input_obj.invalid_keys))
    if not input_obj.input_type:
        message = 'input "{}" does not specify a type. '.format(input_obj.name)
        if input_obj.default:
            if isinstance(input_obj.default, dict):
                for key in input_obj.default.keys():
                    if key in INTRINSIC_FNS:
                        input_obj.default = None
                        if key in STR_INTRINSIC_FNS:
                            message += 'The correct type could be "string".'
                        if key in DICT_INTRINSIC_FNS:
                            message += 'The correct type could be "dict".'
                        if key in INT_INTRINSIC_FNS:
                            message += 'The correct type could be "int".'
                if isinstance(input_obj.default, dict) and not suggestions:
                    message += 'The correct type could be "dict".'
            if get_default_type(input_obj) == str and not suggestions:
                message += 'The correct type could be "string".'
            if get_default_type(input_obj) == bool and not suggestions:
                message += 'The correct type could be "boolean".'
            if isinstance(input_obj.default, list) and not suggestions:
                message += 'The correct type could be "list".'
            if get_default_type(input_obj) == int and not suggestions:
                message += 'The correct type could be "integer".'
            if get_default_type(input_obj) == float and not suggestions:
                message += 'The correct type could be "float".'
        yield LintProblem(line, None, message)
    elif get_type_name(input_obj) not in constants.INPUTS_BY_DSL.get(dsl, []):
        input_type = item[LEVEL1]
        yield LintProblem(
            line,
            None,
            'Input of type {} is not supported by DSL {}.'.format(
                get_type_name(input_obj), dsl
            ),
            start_mark=input_type.start_mark.line,
            end_mark=input_type.end_mark.line
        )
    elif not input_obj.display_label and dsl != 'cloudify_dsl_1_3':
        yield LintProblem(
            line,
            None,
            'Input {} is missing a display_label.'.format(input_obj.name),
            fixable=True
        )
    elif input_obj.display_label and dsl == 'cloudify_dsl_1_3':
        yield LintProblem(
            line,
            None,
            'Display_label is not supported by DSL {}.'.format(dsl)
        )
    elif input_obj.display_label and dsl != 'cloudify_dsl_1_5' and (
            input_obj.display_label in constants.DSL_1_5):
        yield LintProblem(
            line,
            None,
            '{label} is not supported by DSL {dsl}.'.format(
                label=input_obj.display_label, dsl=dsl)
        )
    elif get_type(input_obj) and input_obj.default:
        message = ''
        if isinstance(input_obj.default, dict):
            for key in input_obj.default.keys():
                if key in INTRINSIC_FNS:
                    message = "intrinsic function"
                    if key in STR_INTRINSIC_FNS and not isinstance(
                            get_type_name(input_obj), str):
                        message = 'input "{}" specify a type {}, The correct'\
                            ' type is "string".'.format(
                                input_obj.name,
                                get_type_name(input_obj))
                    if key in INT_INTRINSIC_FNS and not\
                        isinstance(
                                   get_type(input_obj),
                                   int):
                        message = 'input "{}" specify a type {}, The correct'\
                            ' type is "int".'.format(
                                input_obj.name,
                                get_type_name(input_obj))
        if not (message or get_default_type(input_obj) == get_type(input_obj)):
            message = 'input "{}" specify a type {}, However this'\
                ' doesn\'t match deafult of type {}.'.format(
                    input_obj.name,
                    get_type_name(input_obj),
                    input_obj.default)
        if message and message not in ["intrinsic function"]:
            yield LintProblem(line, None, message)
    elif not input_obj.display_label:
        yield LintProblem(
            line,
            None,
            'Input {} is missing a display_label.'.format(input_obj.name),
            fixable=True
        )


def get_type_name(input_obj):
    if isinstance(input_obj.input_type, yaml.nodes.ScalarNode):
        return input_obj.input_type.value
    else:
        raise TypeError()


def get_type(input_obj):
    type_name = get_type_name(input_obj)
    if type_name == 'string':
        return str
    elif type_name == 'integer':
        return int
    elif type_name == 'boolean':
        return bool
    else:
        return globals().get('__builtins__').get(type_name)


def get_default_type(input_obj):
    default_value = input_obj.default
    if type(default_value) == str:  # noqa: E721
        if default_value.isdigit():
            result = int
        elif is_float(default_value):
            result = float
        else:
            result = str
    elif type(default_value) == bool:  # noqa: E721
        result = bool
    else:
        result = type(input_obj.default)
    return result


def is_float(string):
    try:
        float(string)
        return True
    except ValueError:
        return False


class CfyInput(object):
    def __init__(self, nodes):
        self._line = None
        self.name, self.mapping = self.get_input(nodes)
        self.invalid_keys = []
        if self.name and self.mapping:
            for key in list(self.mapping.keys()):
                if key not in ALLOWED_KEYS:
                    self.invalid_keys.append(key)
                    del self.mapping[key]
            self.input_type = self.mapping.get('type')
            self.description = self.mapping.get('description')
            self._default = self.mapping.get('default')
            self.constraints = self.mapping.get('constraints')
            self.display_label = self.mapping.get('display_label')

    @property
    def default(self):
        return self._default

    @default.setter
    def default(self, value):
        self._default = value

    @property
    def line(self):
        return self._line

    def not_input(self):
        return all([not k for k in self.mapping.values()])

    def __dict__(self):
        return {
            self.name: self.mapping
        }

    def get_input(self, nodes):
        if not isinstance(nodes, tuple) or len(nodes) != 2:
            name = None
            mapping = None
        else:
            name = self.get_input_name(nodes[0])
            mapping = self.get_input_mapping(nodes[1])
        return name, mapping

    def get_input_name(self, node):
        if isinstance(node, yaml.nodes.ScalarNode):
            self._line = node.end_mark.line + 1
            return node.value

    def get_input_mapping(self, node):
        mapping = {
            'type': None,
            'default': None,
            'description': None,
            'constraints': None,
            'display_label': None,
        }
        if isinstance(node, yaml.nodes.MappingNode):
            for tup in node.value:
                if not len(tup) == 2:
                    continue
                mapping_name = tup[0].value
                mapping_value = self.get_mapping_value(
                    mapping_name, tup[1])
                mapping[mapping_name] = mapping_value
        return mapping

    def get_mapping_value(self, name, value):
        if name not in ['default', 'constraints']:
            return value
        else:
            return recurse_mapping(value)



########
# Copyright (c) 2014-2023 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

CONCAT = 'concat'
GET_INPUT = 'get_input'
GET_SECRET = 'get_secret'
GET_PROPERTY = 'get_property'
GET_ATTRIBUTE = 'get_attribute'
GET_CAPABILITY = 'get_capability'
GET_ENVIRONMENT_CAPABILITY = 'get_environment_capability'
INSTRINSIC_FUNCTIONS = [
    CONCAT,
    GET_INPUT,
    GET_SECRET,
    GET_PROPERTY,
    GET_ATTRIBUTE,
    GET_CAPABILITY,
    GET_ENVIRONMENT_CAPABILITY
]



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from cfy_lint.yamllint_ext.autofix.utils import filelines
from cfy_lint.yamllint_ext.autofix.indentation.utils import (
    get_yaml_dict,
    get_file_content,
    filter_corrections,
    get_compare_file_content,
    indentify_indentation_corrections
)


def fix_indentation(problem):
    if problem.rule == 'indentation':
        with filelines(problem.file) as lines:
            original = get_file_content(problem.file)
            compare = get_compare_file_content(get_yaml_dict(problem.file))
            corrections = filter_corrections(
                indentify_indentation_corrections(original, compare),
                problem.line)
            for line, correction in sorted(corrections.items()):
                if line == -1:
                    print('Unable to autofix indentation for line {}. '
                          'Unsupported YAML.'.format(problem.line))
                    sys.exit(1)
                lines[line - 1] = correction['new']
            problem.fixed = True



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import re
import sys
import json
import time
import yaml
import pathlib
import urllib.request
from urllib.parse import urlparse
from packaging.version import parse as version_parse

from yamllint.config import YamlLintConfigError

from cfy_lint.yamllint_ext.cloudify.models import NodeTemplate
from cfy_lint.yamllint_ext.constants import (
    UNUSED_IMPORT,
    UNUSED_INPUTS,
    DEFAULT_TYPES,
    BLUEPRINT_MODEL,
    UNUSED_IMPORT_CTX,
    LATEST_PLUGIN_YAMLS,
    NODE_TEMPLATE_MODEL)

INTRINSIC_FNS = [
    'merge',
    'concat',
    'get_sys',
    'get_input',
    'get_label',
    'get_secret',
    'string_find',
    'string_split',
    'string_lower',
    'string_upper',
    'get_property',
    'get_attribute',
    'string_replace',
    'get_capability',
    'get_environment_capability',
]

context = {
    'imports': [],
    'dsl_version': None,
    'inputs': {},
    'imported_node_types': [],
    UNUSED_INPUTS: {},
    UNUSED_IMPORT_CTX: {},
    'node_templates': {},
    'node_types': {},
    'capabilities': {},
    'outputs': {},
    'current_tokens_line': 0,
    'add_label': [],
    'line_diff': {},
}


def assign_current_top_level(elem):
    if isinstance(elem.curr, yaml.tokens.ScalarToken) and \
            elem.curr.value in BLUEPRINT_MODEL and \
            isinstance(elem.nextnext,
                       yaml.tokens.BlockMappingStartToken):
        return elem.curr.value
    elif isinstance(elem.curr, yaml.tokens.BlockEndToken) and \
            isinstance(elem.nextnext, yaml.tokens.ScalarToken) and \
            elem.nextnext.value in BLUEPRINT_MODEL:
        return ''


def assign_nested_node_template_level(elem):
    if not isinstance(elem.curr, yaml.tokens.ScalarToken):
        return
    if elem.curr.value not in NODE_TEMPLATE_MODEL:
        return
    if isinstance(elem.nextnext, (yaml.tokens.BlockMappingStartToken,
                                  yaml.tokens.BlockEntryToken)):
        return elem.curr.value


def update_model(_elem):
    """Tracking a Cloudify Model inside YAMLLINT context.

    :param _elem:
    :return:
    """

    context['current_tokens_line'] = _elem.line_no
    if stop_document(_elem):
        # The document is finished.
        return
    # We are in the middle of the document.
    top_level = assign_current_top_level(_elem)
    node_template(_elem)
    if skip_inputs_in_node_templates(_elem):
        return
    elif isinstance(top_level, str):
        context['current_top_level'] = top_level  # noqa


def stop_document(_elem):
    if isinstance(_elem.curr, yaml.tokens.StreamStartToken):
        # This is the start of the YAML document.
        context['model'] = BLUEPRINT_MODEL
        context['current_top_level'] = None  # noqa
    elif isinstance(_elem.curr, yaml.tokens.StreamEndToken):
        # This is the end of the YAML document.
        del context['model']
        return True
    return False


def node_template(_elem):
    if context.get('current_top_level') == 'node_templates':
        # When we are looking at Node Templates, we may
        nt = assign_nested_node_template_level(_elem)
        if isinstance(nt, str):
            context['node_template_level'] = nt
    else:
        context['node_template_level'] = None


def skip_inputs_in_node_templates(top_level):
    return context.get('current_top_level') == 'node_templates' and \
           top_level == 'inputs'


def get_json_from_marketplace(url):
    try:
        resp = urllib.request.urlopen(url)
    except (urllib.error.HTTPError, urllib.error.URLError) as e:
        if isinstance(e, urllib.error.URLError):
            raise urllib.error.URLError("Connection Error")
        return {}
    body = resp.read()
    return json.loads(body)


def get_plugin_id_from_marketplace(plugin_name):
    url_plugin_id = 'https://marketplace.cloudify.co/plugins?name={}'.format(
        plugin_name)
    json_resp = get_json_from_marketplace(url_plugin_id)
    if 'items' in json_resp:
        if len(json_resp['items']) == 1:
            return json_resp['items'][0]['id']


def get_plugin_versions_from_marketplace(plugin_id):
    url_plugin_version = 'https://marketplace.cloudify.co/' \
                         'plugins/{}/versions?'.format(plugin_id)
    json_resp = get_json_from_marketplace(url_plugin_version)
    if 'items' in json_resp:
        versions = [item['version'] for item in json_resp['items']]
        return sorted(versions, key=lambda x: version_parse(x))
    return []


def get_plugin_release_spec_from_marketplace(plugin_id, plugin_version):
    release_url = 'https://marketplace.cloudify.co/plugins/{}/{}'.format(
        plugin_id, plugin_version)
    return get_json_from_marketplace(release_url)


def validate_versions(versions, validations):
    for neq in validations['!=']:
        if neq in versions:
            versions.remove(neq)
    for gteq in validations['>=']:
        parsed_gteq = version_parse(gteq)
        versions = [v for v in versions if version_parse(v) >= parsed_gteq]
    for lteq in validations['<=']:
        parsed_lteq = version_parse(lteq)
        versions = [v for v in versions if version_parse(v) <= parsed_lteq]
    for gt in validations['>']:
        parsed_gt = version_parse(gt)
        versions = [v for v in versions if version_parse(v) > parsed_gt]
    for lt in validations['<=']:
        parsed_lt = version_parse(lt)
        versions = [v for v in versions if version_parse(v) < parsed_lt]
    return versions


def get_validations(version_constraints):
    validations = {
        '==': [],
        '!=': [],
        '>=': [],
        '<=': [],
        '>': [],
        '<': [],
    }
    # Organize the version constraints so we get a dict like this:
    # {
    #    '==': [],
    #    '!=': ['1.0'],
    #    '>=': ['0.8', 0.9'],
    #    '<=': ['1.1'],
    # }
    try:
        for version_constraint in version_constraints:
            sign = re.match('[\\<\\>\\=\\!]+', version_constraint).group(0)
            plugin_version = re.findall(
                '(\\d+.\\d+.\\d+)', version_constraint)[0]
            validations[sign].append(plugin_version)
    except Exception as e:
        raise YamlLintConfigError('invalid version: %s' % e)
    return validations


def get_version_constraints(plugin_name, plugin_version_string):
    version_constraints = list(
        # Get rid of irrelevant stringy stuff.
        filter(
            lambda item: item, re.split(
                'plugin:| |{}|,'.format(plugin_name),
                plugin_version_string)
        )
    )
    # re.split is afraid of this one.
    if '?version' in version_constraints:
        version_constraints.remove('?version')
    if 'version=' in version_constraints:
        version_constraints.remove('version=')
    return version_constraints


def get_plugin_spec(plugin_version_string, plugin_name):

    version_constraints = get_version_constraints(
        plugin_name, plugin_version_string)

    validations = get_validations(version_constraints)

    plugin_id = get_plugin_id_from_marketplace(plugin_name)
    versions = get_plugin_versions_from_marketplace(plugin_id)

    if len(validations['==']) == 1 and validations['=='][0] in versions:
        return get_plugin_release_spec_from_marketplace(
            plugin_id, validations['=='][0])

    versions = validate_versions(versions, validations)

    if len(versions):
        return get_plugin_release_spec_from_marketplace(
            plugin_id, versions[-1])


def get_plugin_yaml_url(plugin_import):
    plugin_name, plugin_spec = _get_plugin_spec(plugin_import)
    if not plugin_spec:
        return LATEST_PLUGIN_YAMLS.get(plugin_name)
    elif len(plugin_spec.get('yaml_urls', [])):
        return plugin_spec['yaml_urls'][0]['url']


def _get_plugin_spec(plugin_import):
    parsed_import_item = urlparse(plugin_import)
    plugin_name = parsed_import_item.path
    return plugin_name, get_plugin_spec(parsed_import_item.query, plugin_name)


def get_node_types_for_plugin_import(plugin_import):
    plugin_name, plugin_spec = _get_plugin_spec(plugin_import)
    plugin_version = None
    if plugin_spec:
        plugin_version = plugin_spec['version']
    return get_node_types_for_plugin_version(plugin_name, plugin_version)


def get_node_types_for_plugin_version(plugin_name, plugin_version):

    node_types = {}
    offset = 0
    url = 'https://marketplace.cloudify.co/node-types?' \
          '&plugin_name={}' \
          '&plugin_version={}' \
          '&offset={}'.format(plugin_name, plugin_version, offset)

    while True:
        result = get_json_from_marketplace(url)
        if 'items' in result:
            for item in result['items']:
                if item['type'] not in node_types:
                    node_types[item['type']] = item
            # Stop paginating results when the offset has been incrimented
            # Beyond the amount of total reported results.
            if result['pagination']['total'] <= offset:
                break
            offset += 100
            url = re.sub(r'offset=\d+', 'offset={}'.format(offset), url)

    return node_types


def import_cloudify_yaml(import_item, base_path=None, cache_ttl=None):
    cache_ttl = cache_ttl or 86400
    cache_item = re.sub('[^0-9a-zA-Z]+', '_', import_item)
    current_dir = pathlib.Path(__file__).parent.resolve()
    cache_dir = pathlib.Path(os.path.join(
        current_dir,
        'cloudify/__cfylint_runtime_cache')).resolve()
    cache_dir
    if not cache_dir.exists():
        os.makedirs(cache_dir.absolute())
    cache_item_path = pathlib.Path(
        os.path.join(cache_dir.absolute(), cache_item))
    if cache_item_path.exists():
        # Check if the file has been stale for a while
        if cache_item_path.stat().st_ctime + cache_ttl < time.time():
            os.remove(cache_item_path)

    result = {}
    parsed_import_item = urlparse(import_item)
    if parsed_import_item.scheme == 'plugin':
        if cache_item_path.exists():
            with open(cache_item_path.absolute(), 'r') as jsonfile:
                result['node_types'] = json.load(jsonfile)
        else:
            node_types = get_node_types_for_plugin_import(
                import_item)
            result['node_types'] = node_types
            with open(cache_item_path.absolute(), 'w') as jsonfile:
                json.dump(node_types, jsonfile)
        # This is kind of wasteful, but
        # what this does is it stores the node types also
        # per plugin import line.
        # this enables us to analyze
        # if a plugin is being used.
        result[UNUSED_IMPORT] = {
            import_item: list(result['node_types'].keys())
        }
    if parsed_import_item.scheme in ['http', 'https']:
        if cache_item_path.exists():
            with open(cache_item_path.absolute(), 'r') as jsonfile:
                result = json.load(jsonfile)
        else:
            page = urllib.request.Request(
                import_item,
                headers={'User-Agent': 'Mozilla/5.0'}
            )
            try:
                infile = urllib.request.urlopen(page).read()
            except urllib.error.HTTPError:
                print('Error: Unable to reach URL: {}'.format(import_item))
                sys.exit(1)
            result = yaml.safe_load(infile)
            with open(cache_item_path.absolute(), 'w') as jsonfile:
                json.dump(result, jsonfile)
    elif import_item == 'cloudify/types/types.yaml':
        result = DEFAULT_TYPES
    elif base_path and os.path.exists(os.path.join(base_path, import_item)):
        with open(os.path.join(base_path, import_item), 'r') as stream:
            result = yaml.safe_load(stream)
            node_types_used = make_list_types(result)
            delete_imports_from_unused_ctx(node_types_used)
            add_to_imported_node_types(node_types_used)

    elif os.path.exists(import_item):
        with open(import_item, 'r') as stream:
            result = yaml.safe_load(stream)
        result = result or {}

    for k in result.keys():
        left = 'imported_{}'.format(k)
        if left not in context:
            if isinstance(result[k], dict) and left in ['imported_node_types']:
                context[left] = list(result[k].keys())
            else:
                context[left] = result[k]
        elif isinstance(context[left], list):
            if k in ["tosca_definitions_version"]:
                context[left].append(result[k])
            else:
                context[left].extend(result[k])
        elif isinstance(context[left], str):
            if context[left] != result[k] and \
                    k in ["tosca_definitions_version"]:
                if not isinstance(context[left], list):
                    tmp = [context[left]]
                    context[left] = tmp
                context[left].append(result[k])
            elif context[left] != result[k]:
                raise Exception(
                    'There is no match between '
                    '{result} and {context}'.format(
                        context=context[left],
                        result=result[k]))
        else:
            context[left].update(result[k])


def delete_imports_from_unused_ctx(node_types_used):
    need_to_del = []
    unused_import = context[UNUSED_IMPORT_CTX].keys()
    for type in node_types_used:
        for k in unused_import:
            if type in context[UNUSED_IMPORT_CTX][k]:
                need_to_del.append(k)

    for d in need_to_del:
        del context[UNUSED_IMPORT_CTX][d]

    if 'plugin:cloudify-fabric-plugin' in context[UNUSED_IMPORT_CTX].keys():
        del context[UNUSED_IMPORT_CTX]['plugin:cloudify-fabric-plugin']


def make_list_types(content_file):
    values = []
    keys = ['type', 'derived_from']
    for k, v in content_file.items():
        if 'node_templates' == k:
            values.extend(find_values_by_key(v, keys))
        if 'node_types' == k:
            values.extend(v.keys())
            values.extend(find_values_by_key(v, ['derived_from']))
    return values


def find_values_by_key(yaml_data, keys):
    """
    Find all values associated with a given key in YAML data.
    :param yaml_data: YAML data to search through.
    :param keys: List of keys to look for in the YAML data.
    :return: List of all values associated with the given key.
    """
    values = []
    if isinstance(yaml_data, dict):
        for k, v in yaml_data.items():
            if k in keys:
                values.append(v)
            elif isinstance(v, (dict, list)):
                nested_values = find_values_by_key(v, keys)
                if nested_values:
                    values.extend(nested_values)
    elif isinstance(yaml_data, list):
        for i in yaml_data:
            if isinstance(i, (dict, list)):
                nested_values = find_values_by_key(i, keys)
                if nested_values:
                    values.extend(nested_values)
            elif i in keys:
                values.append(i)
    return values


def setup_types(buffer=None, data=None, base_path=None):
    try:
        data = data or yaml.safe_load(buffer)
    except yaml.parser.ParserError:
        return
    if not data:
        return
    for imported in data.get('imports', {}):
        import_cloudify_yaml(imported, base_path=base_path)
    add_to_node_types(data.get('node_types', {}))


def add_to_imported_node_types(node_types_used):
    if isinstance(node_types_used, list):
        for item in node_types_used:
            if item not in context['imported_node_types']:
                context['imported_node_types'].append(item)
    elif node_types_used not in context['imported_node_types']:
        context['imported_node_types'].append(node_types_used)


def add_to_node_types(node_types):
    context['imported_node_types'].extend(node_types.keys())


def setup_node_templates(elem):
    if 'node_templates' not in context:
        context['node_templates'] = {}
    if elem.prev and elem.prev.node.value == 'node_templates':
        for item in elem.node.value:
            node_template = setup_node_template(item)
            if node_template.name not in context:
                context['node_templates'].update({
                    node_template.name: node_template
                })
    elem.node_templates = context['node_templates']


def setup_node_template(list_item):
    if len(list_item) == 2:
        if isinstance(list_item[0], yaml.nodes.ScalarNode) and \
                isinstance(list_item[1], yaml.nodes.MappingNode):
            node_template = NodeTemplate(list_item[0].value)
            node_template.node_type = setup_node_type(list_item[1].value)
            return node_template


def setup_node_type(value):
    return value[0][1].value


def mapping_is_two_length_intrinsic_function(mapping):
    if len(mapping) == 2 and not isinstance(mapping[0], tuple):
        try:
            if mapping[0].value in INTRINSIC_FNS:
                return True
        except AttributeError:
            return False


def mapping_is_one_length_intrinsic_function_tuple(mapping):
    if len(mapping) == 1 and isinstance(mapping[0], tuple):
        if len(mapping[0]) == 2 and mapping[0][0].value in INTRINSIC_FNS:
            return True


def mapping_is_one_length_intrisic_function_mapping_node(mapping):
    if len(mapping) == 1 and isinstance(mapping[0],
                                        yaml.nodes.MappingNode):
        try:
            if len(mapping[0].value) == 2 and \
                   mapping[0].value[0].value in INTRINSIC_FNS:
                return True
        except AttributeError:
            return False


def recurse_mapping(mapping):
    if isinstance(mapping, dict):
        new_dict = {}
        for k, v in mapping.items():
            new_dict[k] = recurse_mapping(v)
        return new_dict
    elif isinstance(mapping, (list, tuple)):
        new_list = []
        if mapping_is_two_length_intrinsic_function(mapping):
            return recurse_mapping({mapping[0].value: mapping[1].value})
        if mapping_is_one_length_intrinsic_function_tuple(mapping):
            return recurse_mapping(
                {
                    mapping[0][0].value: mapping[0][1].value
                }
            )
        if mapping_is_one_length_intrisic_function_mapping_node(mapping):
            return recurse_mapping(
                {
                    mapping[0].value[0].value: mapping[0].value[1].value
                }
            )
        for item in mapping:
            new_list.append(recurse_mapping(item))
        return new_list
    elif not isinstance(mapping, yaml.nodes.Node):
        return mapping
    elif isinstance(mapping, yaml.nodes.ScalarNode):
        if 'bool' in mapping.tag:
            return bool(mapping.value.lower() == "true")
        return mapping.value
    elif isinstance(mapping, yaml.nodes.SequenceNode):
        new_list = []
        for item in mapping.value:
            new_list.append(recurse_mapping(item))
        return new_list
    elif isinstance(mapping, yaml.nodes.MappingNode):
        new_dict = {}
        new_list = []
        for item in mapping.value:
            if isinstance(item, (list, tuple)) and len(item) == 2:
                key = item[0].value
                value = recurse_mapping(item[1].value)
                new_dict[key] = value
            else:
                new_list.append(item)
        if new_dict:
            return new_dict
        return new_list


def process_relevant_tokens(model, keyword):
    def wrapper_outer(function):
        def wrapper_inner(*args, **kwargs):
            token = kwargs.get('token')
            if isinstance(token, model):
                if isinstance(keyword, str):
                    if token.prev and token.prev.node.value == keyword:
                        yield from function(*args, **kwargs)
                if isinstance(keyword, list):
                    if token.prev and token.prev.node.value in keyword:
                        yield from function(*args, **kwargs)
        return wrapper_inner
    return wrapper_outer


def update_dict_values_recursive(default_dict, name_file_config):
    with io.open(name_file_config):
        f = open("config.yaml", "r")
        user_dict = f.read()

    default_dict = yaml.load(default_dict)
    user_dict = yaml.load(user_dict)

    if user_dict and default_dict:
        for key, value in user_dict.items():
            if value is dict:
                update_dict_values_recursive(default_dict[key], value)
            if value:
                default_dict[key] = value
    return default_dict


def check_node_imported(node_name):
    for keys in context['imported_node_types_by_plugin']:
        if node_name in context['imported_node_types_by_plugin'][keys]:
            return True
    return False


def recurse_get_readable_object(mapping):
    if isinstance(mapping, yaml.nodes.ScalarNode):
        return mapping.value
    if isinstance(mapping, yaml.nodes.MappingNode):
        mapping_list = []
        for item in mapping.value:
            mapping_list.append(recurse_get_readable_object(item))
        mapping_dict = {}
        for item in mapping_list:
            try:
                mapping_dict[item[0]] = item[1]
            except KeyError:
                mapping_dict.update(item)
        return mapping_dict
    elif isinstance(mapping, tuple):
        if len(mapping) == 2 and isinstance(mapping[0], yaml.nodes.ScalarNode):
            return {
                mapping[0].value: recurse_get_readable_object(mapping[1])
            }
        else:
            new_list = []
            for item in mapping:
                new_list.append(recurse_get_readable_object(item))
            return new_list
    elif isinstance(mapping, yaml.nodes.SequenceNode):
        new_list = []
        for item in mapping.value:
            new_list.append(recurse_get_readable_object(item))
        return new_list


def add_severity(problem):
    if problem.rule in ['empty-lines', 'colons', 'brackets',
                        'commas', 'trailing-spaces']:
        problem.severity = 0
    elif problem.rule in ['capabilities', 'truthy', 'line-length'] or \
        (problem.rule == 'inputs' and 'missing a display_label' in
         problem.message or 'does not specify a type' in problem.message) or \
            (problem.rule == 'node_templates' and
                             'does not provide Tags' in problem.message):
        problem.severity = 1
    elif problem.rule in ['node_templates', 'indentation',
                          'relationships'] or \
            problem.rule == 'inputs' and 'unused' in problem.message:
        problem.severity = 2
    elif problem.rule in ['dsl_version', 'inputs',
                          'node_templates', 'empty-values']:
        problem.severity = 4
    else:
        problem.severity = None



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cfy_lint.yamllint_ext.autofix.utils import filelines, get_eol


def fix_deprecated_relationships(problem):
    if problem.rule == 'relationships' and \
            'deprecated relationship type' in problem.message:
        with filelines(problem.file) as lines:
            line = lines[problem.line]
            line, eol = get_eol(line)
            split = problem.message.split()
            new_line = line.replace(split[-5], split[-3].rstrip('.'))
            lines[problem.line] = new_line + eol
        problem.fixed = True



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from cfy_lint.yamllint_ext.utils import context
from cfy_lint.yamllint_ext.autofix.utils import filelines

PATTERN = "^ *\n"


def fix_empty_lines(problem):
    if problem.fix_all or problem.fix_new_lines:
        with filelines(problem.file) as lines:
            successive_blank_lines = 0
            deleted_lines = 0
            index = 0
            current_sum = 0
            keys = []
            context['line_diff'][0] = 0

            # remove blanklines from start of file
            while re.match(PATTERN, lines[0]):
                lines.pop(0)

            while index < (len(lines) - 1):
                line = lines.pop(index)

                if re.match(PATTERN, line):
                    successive_blank_lines -= 1
                    if successive_blank_lines <= -2:
                        continue
                else:
                    if successive_blank_lines < -1:
                        current_sum += successive_blank_lines + 1
                        context['line_diff'][index + deleted_lines] = \
                            current_sum
                        deleted_lines -= successive_blank_lines + 1
                        keys.append(index)
                    successive_blank_lines = 0

                lines.insert(index, line)
                index += 1

            # remove blanklines from end of file
            while re.match(PATTERN, lines[-1]):
                lines.pop(-1)



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import io
from .. import generators


def get_buffer(file_path='resources/blueprint.yaml', read=False):
    pp = os.path.join(os.path.dirname(__file__), file_path)
    return io.open(pp, newline='')


def get_gen(file_path='resources/blueprint.yaml',
            gen=generators.node_generator):
    buffer = get_buffer(file_path)
    return gen(buffer)


def get_file_obj(content):
    return io.StringIO(content, newline='')


def get_loader(yaml_content):
    buffer = get_file_obj(yaml_content)
    return generators.SafeLineLoader(buffer)


def get_gen_as_list(callable, payload):
    while True:
        try:
            if isinstance(payload, dict):
                return list(
                    callable(
                        **payload
                    )
                )
            else:
                return list(
                    callable(
                        payload
                    )
                )

        except (StopIteration, AttributeError):
            break



########
# Copyright (c) 2014-2022 Cloudify Platform Ltd. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys
import pathlib
from setuptools import (setup, find_packages)


def get_version():
    current_dir = pathlib.Path(__file__).parent.resolve()

    with open(os.path.join(current_dir,
                           'cfy_lint/__version__.py'),
              'r') as outfile:
        var = outfile.read()
        return re.search(r'\d+.\d+.\d+', var).group()


install_requires = [
    'click>8,<9',
    'yamllint==1.28.0',
    'packaging>=17.1,<=21.3',
]

if sys.version_info.major == 3 and sys.version_info.minor == 6:
    install_requires += [
        'pyyaml>=5.4.1',
        'networkx>=1.9.1,<=3.1'
    ]
else:
    install_requires += [
        'pyyaml>=6.0.1',
        'networkx>=3.2.1'
    ]


setup(
    name='cfy-lint',
    version=get_version(),
    license='LICENSE',
    packages=find_packages(),
    description='Linter for Cloudify Blueprints',
    entry_points={
        "console_scripts": [
            "cfy-lint = cfy_lint.main:lint",
        ]
    },
    package_data={
        'cfy_lint': [
            'yamllint_ext/cloudify/__cfylint_runtime_cache/README.md',
        ]
    },
    install_requires=install_requires
)
